RDF: str = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
RDF_TYPE: str = f'{RDF}type'
RDF_VALUE: str = f'{RDF}value'
RDF_LIST: str = f'{RDF}List'
RDF_FIRST: str = f'{RDF}first'
RDF_REST: str = f'{RDF}rest'
RDF_NIL: str = f'{RDF}nil'
RDF_DIRECTION: str = f'{RDF}direction'
RDF_LANGUAGE: str = f'{RDF}language'
RDF_JSON: str = f'{RDF}JSON'
RDF_LANGSTRING: str = f'{RDF}langString'

XSD: str = 'http://www.w3.org/2001/XMLSchema#'
XSD_BOOLEAN: str = f'{XSD}boolean'
XSD_DECIMAL: str = f'{XSD}decimal'
XSD_DOUBLE: str = f'{XSD}double'
XSD_INTEGER: str = f'{XSD}integer'
XSD_STRING: str = f'{XSD}string'

I18N: str = 'https://www.w3.org/ns/i18n#'
